#!/usr/bin/python3
#
# SPDX-License-Identifier: GPL-2.0-only
# (c) 2026 Gerd Hoffmann
#
# pylint: disable=too-many-arguments
# pylint: disable=too-many-positional-arguments
# pylint: disable=too-many-public-methods
""" manage qemu firmware config files """
import os
import json
import subprocess

class QemuFirmware:
    """ manage one qemu firmware config file """

    def __init__(self, data = None, filename = None):
        self.data = data
        if self.data is None:
            self.data = {}
        if filename:
            with open(filename, 'r', encoding = 'utf-8') as f:
                self.data = json.loads(f.read())


    def cmdline_flash(self, name, prop, desc, readonly = False, snapshot = False):
        filename = desc.get('filename')
        fformat = desc.get('format')
        drive = f'if=none,id={name},format={fformat},file={filename}'
        if readonly:
            drive += ',readonly=on'
        if snapshot:
            drive += ',snapshot=on'
        cmdline = [ '-drive', drive,
                    '-machine', f'{prop}={name}' ]
        return cmdline

    def cmdline_igvm(self, igvm):
        cmdline = []
        cmdline += [ '-object', f'igvm-cfg,id=igvm0,file={igvm}' ]
        cmdline += [ '-machine', 'igvm-cfg=igvm0' ]
        return cmdline

    def cmdline(self, varsfile = None):
        cmdline = []
        mapping = self.data.get('mapping')
        device = mapping.get('device')
        mode = mapping.get('mode')
        fw_code = mapping.get('executable')
        if device == 'flash' and mode == 'split':
            fw_vars = mapping.get('nvram-template').copy()
            if varsfile:
                fw_vars['filename'] = varsfile
                snapshot = False
            else:
                snapshot = True
            cmdline += self.cmdline_flash('code', 'pflash0', fw_code, readonly = True)
            cmdline += self.cmdline_flash('vars', 'pflash1', fw_vars, snapshot = snapshot)
        elif device == 'flash' and mode == 'stateless':
            cmdline += self.cmdline_flash('code', 'pflash0', fw_code, readonly = True)
        elif device == 'memory':
            cmdline += [ '-bios', mapping.get('filename') ]
        elif device == 'igvm':
            cmdline += self.cmdline_igvm(mapping.get('filename'))
        else:
            raise RuntimeError(f'not supported: device={device} mode={mode}')

        if self.requires_smm():
            cmdline += [ '-machine', 'smm=on' ]
            cmdline += [ '-global', 'driver=cfi.pflash01,property=secure,value=on']
        if self.requires_qemu_vars():
            if self.arch() == 'x86_64':
                device = 'uefi-vars-x64'
            else:
                device = 'uefi-vars-sysbus'
            if varsfile:
                device += f',jsonfile={varsfile}'
            cmdline += [ '-device', device ]

        return cmdline

    def vars_template(self):
        mapping = self.data.get('mapping')
        if mapping.get('device') == 'flash':
            if mapping.get('mode') != 'split':
                return None
            return mapping.get('nvram-template').get('filename')
        if mapping.get('device') == 'memory':
            uefivars = mapping.get('uefi-vars')
            if uefivars:
                return uefivars.get('template')
        return None

    def arch(self):
        targets = self.data.get('targets')[0]
        return targets.get('architecture')

    def is_native(self):
        hostarch = os.uname().machine
        return hostarch == self.arch()

    def machine(self):
        targets = self.data.get('targets')[0]

        if len(targets.get('machines')) == 0:
            if self.arch() == 'x86_64':
                return 'q35'
            return 'virt'

        machine = targets.get('machines')[0]
        if 'pc-i440fx' in machine:
            return 'pc'
        if 'pc-q35' in machine:
            return 'q35'
        if 'virt' in machine:
            return 'virt'
        return machine


    def has_feature(self, name):
        features = self.data.get('features')
        if features is None:
            return False
        return name in features

    def has_sb_enabled(self):
        return self.has_feature('enrolled-keys')

    def has_cc_sev(self):
        return self.has_feature('amd-sev')

    def has_cc_snp(self):
        return self.has_feature('amd-sev-snp')

    def has_cc_tdx(self):
        return self.has_feature('intel-tdx')

    def has_cc(self):
        return self.has_cc_sev() or self.has_cc_snp() or self.has_cc_tdx()

    def requires_smm(self):
        return self.has_feature('requires-smm')

    def requires_qemu_vars(self):
        return self.has_feature('host-uefi-vars')

    def requires_igvm(self):
        device = self.data.get('mapping').get('device')
        return device == 'igvm'

    def check_qemu(self, binary):
        if self.requires_igvm():
            objects = subprocess.run([ binary, '-object', 'help'],
                                     capture_output = True,
                                     check = True).stdout.decode().strip()
            if not 'igvm-cfg' in objects:
                return 'qemu does not support igvm'

        if self.requires_qemu_vars():
            devices = subprocess.run([ binary, '-device', 'help'],
                                     capture_output = True,
                                     check = True).stdout.decode().strip()
            if not 'uefi-vars' in devices:
                return 'qemu does not support uefi-vars'

        return None

    def quirk_code_filename(self):
        mapping = self.data.get('mapping')
        fw_code = mapping.get('executable')
        if fw_code:
            return fw_code.get('filename')
        return mapping.get('filename')

    # should be available as feature
    def quirk_has_hashes(self):
        return 'amdsev' in self.quirk_code_filename()


    @staticmethod
    def inspect_ovmf(filename):
        inspect = subprocess.run([ 'virt-fw-dump', '--modules', '-i', filename ],
                                 capture_output = True,
                                 check = True).stdout.decode().strip()
        features = []
        if 'name=SmmAccess2Dxe' in inspect:
            features.append('requires-smm')
        if 'name=SecureBootConfigDxe' in inspect:
            features.append('secure-boot')
        if 'name=AmdSevDxe' in inspect:
            features.append('amd-sev')
            features.append('amd-sev-es')
            features.append('amd-sev-snp')
        if 'name=TdxDxe' in inspect:
            features.append('intel-tdx')
        if 'name=VirtMmCommunication' in inspect:
            features.append('host-uefi-vars')
        return features

    @staticmethod
    def inspect_igvm(filename):
        inspect = subprocess.run([ 'igvm-inspect', filename ],
                                 capture_output = True,
                                 check = True).stdout.decode().strip()
        features = []
        if 'platform: SEV_SNP' in inspect:
            features.append('amd-sev-snp')
        return features


    @staticmethod
    def create_flash_one(filename):
        fw_format = 'raw'
        if filename.endswith('.qcow2'):
            fw_format = 'qcow2'
        entry = {
            'filename' : filename,
            'format'   : fw_format
        }
        return entry

    @staticmethod
    def create_targets(arch, machine):
        machines = []
        if machine:
            machines.append(machine)
        target = {
            'architecture': arch,
            'machines': machines,
        }
        return [ target, ]

    @staticmethod
    def create_flash_split(fw_code, fw_vars, arch, machine):
        mapping = {
            'device'         : 'flash',
            'mode'           : 'split',
            'executable'     : QemuFirmware.create_flash_one(fw_code),
            'nvram-template' : QemuFirmware.create_flash_one(fw_vars),
        }
        targets = QemuFirmware.create_targets(arch, machine)
        features = QemuFirmware.inspect_ovmf(fw_code)
        return QemuFirmware(data = { 'mapping' : mapping,
                                     'targets' : targets,
                                     'features' : features })

    @staticmethod
    def create_flash_stateless(fw_code, arch, machine):
        mapping = {
            'device'         : 'flash',
            'mode'           : 'stateless',
            'executable'     : QemuFirmware.create_flash_one(fw_code),
        }
        targets = QemuFirmware.create_targets(arch, machine)
        features = QemuFirmware.inspect_ovmf(fw_code)
        return QemuFirmware(data = { 'mapping' : mapping,
                                     'targets' : targets,
                                     'features' : features })

    @staticmethod
    def create_memory(fw_code, arch, machine):
        mapping = {
            'device'   : 'memory',
            'filename' : fw_code,
        }
        targets = QemuFirmware.create_targets(arch, machine)
        features = QemuFirmware.inspect_ovmf(fw_code)
        return QemuFirmware(data = { 'mapping' : mapping,
                                     'targets' : targets,
                                     'features' : features })

    @staticmethod
    def create_igvm(fw_igvm, arch, machine):
        mapping = {
            'device'   : 'igvm',
            'filename' : fw_igvm,
        }
        targets = QemuFirmware.create_targets(arch, machine)
        features = QemuFirmware.inspect_igvm(fw_igvm)
        return QemuFirmware(data = { 'mapping'  : mapping,
                                     'targets'  : targets,
                                     'features' : features })

    @staticmethod
    def from_env():
        fw_json = os.environ.get('EDK2_JSON')
        if fw_json:
            return QemuFirmware(filename = fw_json)
        fw_code = os.environ.get('EDK2_CODE')
        fw_vars = os.environ.get('EDK2_VARS')
        fw_igvm = os.environ.get('EDK2_IGVM')
        fw_arch = os.environ.get('EDK2_ARCH', os.uname().machine)
        fw_machine = os.environ.get('EDK2_MACHINE')
        if fw_code and fw_vars:
            return QemuFirmware.create_flash_split(fw_code, fw_vars, fw_arch, fw_machine)
        if fw_code:
            return QemuFirmware.create_memory(fw_code, fw_arch, fw_machine)
        if fw_igvm:
            return QemuFirmware.create_igvm(fw_igvm, fw_arch, fw_machine)
        raise RuntimeError('no firmware config found (in EDK2_* env vars)')



#!/usr/bin/python3
#
# SPDX-License-Identifier: GPL-2.0-only
# (c) 2026 Gerd Hoffmann
#
# pylint: disable=too-many-arguments
# pylint: disable=too-many-positional-arguments
""" generate cloud init iso images """
import pathlib
import tempfile
import subprocess

import yaml

from passlib.hash import sha512_crypt

class CloudInitISO:
    """ cloud init iso image generator """

    def __init__(self, hostname):
        self.config = {
            "hostname": hostname,
            "users": [],
            "packages": [],
            "runcmd": [],
            "bootcmd": [],
        }
        self.meta = {
            "local-hostname": hostname
        }

    def add_user(self, name, passwd = None, ssh_keys = None, sudo = False, shell = "/bin/bash"):
        user = {
            "name"  : name,
            "shell" : shell
        }
        if passwd:
            cpasswd = sha512_crypt.hash(passwd)
            user["passwd"] = cpasswd
            user["lock_passwd"] = False
        if ssh_keys:
            user["ssh_authorized_keys"] = ssh_keys
        if sudo:
            user["sudo"] = ["ALL=(ALL) NOPASSWD:ALL"]
        self.config["users"].append(user)

    def add_package(self, package_name):
        self.config["packages"].append(package_name)

    def add_runcmd(self, command):
        """ command to run on first boot """
        self.config["runcmd"].append(command)

    def add_bootcmd(self, command):
        """ command to run on every boot """
        self.config["bootcmd"].append(command)

    # pylint: disable=consider-using-with
    def create_iso(self):
        iso = tempfile.NamedTemporaryFile(suffix = '.iso')
        with tempfile.TemporaryDirectory() as tmpdir:
            tmpdir = pathlib.Path(tmpdir)

            user_data_file = tmpdir / "user-data"
            with user_data_file.open("w") as f:
                f.write("#cloud-config\n")
                yaml.safe_dump(self.config, f, default_flow_style=False)

            meta_data_file = tmpdir / "meta-data"
            with meta_data_file.open("w") as f:
                yaml.safe_dump(self.meta, f, default_flow_style=False)

            subprocess.run([
                "xorrisofs",
                "-output", iso.name,
                "-volid", "cidata",
                "-joliet",
                "-rock",
                str(user_data_file),
                str(meta_data_file)
            ], capture_output = True, check=True)

        return iso



#!/usr/bin/python3
#
# SPDX-License-Identifier: GPL-2.0-only
# (c) 2023 Gerd Hoffmann
#
""" certificate and signature helper functions """
import struct
import sys
import hashlib
import logging

from cryptography import x509
from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives.serialization import pkcs7

from virt.firmware.efi import guids


if sys.version_info >= (3, 8):
    import importlib.metadata
    cryptography_version = importlib.metadata.version('cryptography')
else:
    from pkg_resources import get_distribution
    cryptography_version = get_distribution('cryptography').version

cryptography_major = int(cryptography_version.split('.')[0])


def cert_common_name(cert):
    try:
        scn = cert.get_attributes_for_oid(x509.oid.NameOID.COMMON_NAME)[0]
        return scn.value
    except IndexError:
        return 'no CN'

def pe_authenticode_hash(pe, method = 'sha256'):
    h = hashlib.new(method)
    blob = pe.__data__

    csum_off = pe.OPTIONAL_HEADER.get_file_offset() + 0x40
    hdr_end = pe.OPTIONAL_HEADER.SizeOfHeaders

    # hash header, excluding checksum and security directory
    h.update(blob [ 0 : csum_off ])
    logging.debug('hash 0x%06x -> 0x%06x - header start -> csum', 0, csum_off)
    if pe.OPTIONAL_HEADER.NumberOfRvaAndSizes < 4:
        sec = None
        h.update(blob [ csum_off + 4 : hdr_end ])
        logging.debug('hash 0x%06x -> 0x%06x - header csum -> end', csum_off + 4, hdr_end)
    else:
        sec = pe.OPTIONAL_HEADER.DATA_DIRECTORY[4]
        sec_off = sec.get_file_offset()
        h.update(blob [ csum_off + 4 : sec_off ])
        h.update(blob [ sec_off + 8 : hdr_end ])
        logging.debug('hash 0x%06x -> 0x%06x - header csum -> sigs', csum_off + 4, sec_off)
        logging.debug('hash 0x%06x -> 0x%06x - header sigs -> end', sec_off + 8, hdr_end)

    # hash sections
    offset = hdr_end
    for section in sorted(pe.sections, key = lambda s: s.PointerToRawData):
        start = section.PointerToRawData
        end = start + section.SizeOfRawData
        name = section.Name.rstrip(b'\0').decode()
        logging.debug('hash 0x%06x -> 0x%06x - section \'%s\'', start, end, name)
        if start != offset:
            logging.error('unexpected section start 0x%06x (expected 0x%06x, section \'%s\')',
                          start, offset, name)
        h.update(blob [ start : end ])
        offset = end

    # hash remaining data
    if sec and sec.Size:
        end = sec.VirtualAddress
    else:
        end = len(blob)
    if offset < end:
        h.update(blob [ offset : end ])
        logging.debug('hash 0x%06x -> 0x%06x - remaining data', offset, end)

    # hash dword padding
    padding = ((end + 7) & ~7) - end
    if padding:
        for i in range(padding):
            h.update(b'\0')
        logging.debug('hash %d padding byte(s)', padding)

    # log signatures and EOF
    if sec and sec.Size:
        start = sec.VirtualAddress
        end = start + sec.Size
        logging.debug('sigs 0x%06x -> 0x%06x', start, end)
    logging.debug('EOF  0x%06x', len(blob))

    return h.digest()

def pe_type2_signatures(pe):
    siglist = []
    sighdr = pe.OPTIONAL_HEADER.DATA_DIRECTORY[4]
    if sighdr.VirtualAddress and sighdr.Size:
        sigs = pe.__data__[ sighdr.VirtualAddress :
                            sighdr.VirtualAddress + sighdr.Size ]
        pos = 0
        while pos + 8 < len(sigs):
            (slen, srev, stype) = struct.unpack_from('<LHH', sigs, pos)
            if stype == 2:
                siglist.append(sigs [ pos + 8 : pos + slen ])
            pos += slen
            pos = (pos + 7) & ~7 # align
    return siglist

def is_cert_in_sigdb(cert, variable):
    if variable is None:
        return False
    for item in variable.sigdb:
        if item.x509:
            if item.x509 == cert:
                return True
    return False

def is_cert_issuer_in_sigdb(cert, variable):
    if variable is None:
        return False
    for item in variable.sigdb:
        if item.x509:
            try:
                cert.verify_directly_issued_by(item.x509)
                return True
            except (ValueError, TypeError):
                pass
    return False

#
# This does only check whenever one of the certificates in the pkcs7
# signature is found in the given efi variable, either the certificate
# itself or the issuer of the certificate.
#
# The pkcs7 signature itself is NOT verified.
#
# This requires cryptography_major >= 40.
#
def pe_check_cert(siglist, variable):
    if not variable:
        return None
    for sig in siglist:
        sigcerts = pkcs7.load_der_pkcs7_certificates(sig)
        for sigcert in sigcerts:
            for dbcert in variable.sigdb:
                if dbcert.x509:
                    try:
                        sigcert.verify_directly_issued_by(dbcert.x509)
                        return sigcert
                    except (ValueError, TypeError):
                        pass
                    except InvalidSignature:
                        pass
                    if sigcert == dbcert.x509:
                        return sigcert
    return None

def pe_check_hash(digest, variable):
    if not variable:
        return False
    return variable.sigdb.has_sig(guids.EfiCertSha256, digest)



#!/usr/bin/python3
#
# SPDX-License-Identifier: GPL-2.0-only
# (c) 2026 Gerd Hoffmann
#
# pylint: disable=too-many-arguments
# pylint: disable=too-many-positional-arguments
# pylint: disable=too-many-branches
""" misc helper functions for automagic configuration """
import os
import glob

from virt.qemutest.firmware import QemuFirmware

QEMU_BINARY = os.environ.get("QEMU_BINARY")

def find_qemu_binary(guestarch = None):
    hostarch = os.uname().machine
    if guestarch is None:
        guestarch = hostarch
    path = os.environ.get('PATH').split(':')
    name = f'qemu-system-{guestarch}'
    bins = []
    for directory in path:
        bins.append(f'{directory}/{name}')
    if guestarch == hostarch:
        bins.append("/usr/bin/qemu-kvm")
        bins.append("/usr/libexec/qemu-kvm")
    if QEMU_BINARY is not None:
        bins.insert(0, QEMU_BINARY)
    for qemu in bins:
        if os.path.exists(qemu):
            return qemu
    return None

def qemu_default_tcg_cpu(arch):
    if arch in ('riscv64', 'loongarch64'):
        return 'max'
    if arch == 'aarch64':
        return 'cortex-a76'
    if arch == 'arm':
        return 'cortex-a15'
    if arch == 'i386':
        return 'coreduo'
    if arch == 'x86_64':
        return 'Haswell'
    raise RuntimeError(f'unknown arch: {arch}')

def qemu_default_machine(arch = None):
    if arch is None:
        arch = os.uname().machine
    if arch in ('arm', 'aarch64', 'riscv64', 'loongarch64'):
        return 'virt'
    if arch == 'i386':
        return 'pc'
    if arch == 'x86_64':
        return 'q35'
    raise RuntimeError(f'unknown arch: {arch}')

def find_firmware(arch = None, machine = None,
                  smm = False, qemuvars = False, sb = False,
                  sev = False, snp = False):
    if arch is None:
        arch = os.uname().machine
    if machine is None:
        machine = qemu_default_machine(arch)
    mbase = machine.split(',')[0]

    for filename in sorted(glob.glob('/usr/share/qemu/firmware/*.json')):
        fw = QemuFirmware(filename = filename)
        if not 'uefi' in fw.data.get('interface-types'):
            continue

        # check sb features
        if fw.requires_smm() and not smm:
            continue
        if qemuvars and not fw.requires_qemu_vars():
            continue
        if sb and not fw.has_sb_enabled():
            continue

        # check cc features
        if sev and not fw.has_cc_sev():
            continue
        if snp and not fw.has_cc_snp():
            continue

        # check arch + machine type
        for target in fw.data.get('targets'):
            if target.get('architecture') != arch:
                continue
            for item in target.get('machines'):
                if mbase == item:
                    return fw
                if mbase == 'virt' and item == 'virt-*':
                    return fw
                if mbase == 'q35' and item == 'pc-q35-*':
                    return fw
                if mbase == 'pc' and item == 'pc-i440fx-*':
                    return fw

    return None
